#ifndef Graph_hh
#define Graph_hh


#include "Array.hh"
#include "Random.hh"
#include <LEDA/string.h>
#include <LEDA/ugraph.h>


/**
	@short Data structure to handle static graphs.
	@version 1.0
	@author Jordi Petit i Silvestre <jpetit@lsi.upc.es>

	Graph is a data structure to handle static graphs.

	Graph constains information on the number of nodes, edges, degrees,
	adjacency lists, etc.

	A longer description should go here!
*/


class Graph
{
	private:
	
		/** Number of nodes	*/
		int _n;				
		/** Number of edges */
		int _m;				
		/** Adjacency Array */
		Array<int> _adj;		
		/** Edge Array */
		Array<int> _edg;		
		/** Degree Array */
		Array<int> _deg;		
		
		/* Does a DFS traversal */
		void dfs (const int u, Array<bool>& vis) const;

		
	public:
		/** Constructor: reads the graph from a file. If filename="" creates an empty graph. */
		Graph (const string& filename="");
		
		/** Input in "gra" format */
		void get_gra (istream& is);		
		/** Input in "gra" format */
		void get_gra (const string& filename);	
		/** Output in "gra" format */
		void put_gra (ostream& os) const;
		/** Output in "gra" format */
		void put_gra (const string& filename) const;
		/** Output in "cha" format */
		void put_cha (ostream& os) const;
		/** Output in "cha" format */
		void put_cha (const string& filename) const;
		/** Output in "rmf" format */
		void put_rmf (ostream& os) const;
		/** Output in "rmf" format */
		void put_rmf (const string& filename) const;

		/** Returns the number of nodes in the graph. */
		inline int nodes () const {return _n;}
		/** Returns the number of edges in the graph. */
		inline int edges () const {return _m;}
		/** Returns the starting adjacency index of a node. */
		inline int adj (const int u) const {return _adj[u];}
		/** Returns the target node of edge e. */
		inline int edg (const int e) const {return _edg[e];}
		/** Returns the degree of a node. */
		inline int deg (const int u) const {return _deg[u];}
	
		/** Answers wether u and v are connected by an edge */
		bool connected (const int u, const int v) const;
		/** Returns the maximum degree of the graph. */
		int max_degree () const;
		/** Returns the number of connected components in the graph. */
		int components () const;
		/** Returns a node from the graph selected uniformly at random. */
		inline int random_node () const {return Random(nodes());}
		/** Returns a a random neighbor of u selected uniformly at random. */
		inline int random_neighbor (const int u) const {return edg(adj(u)+Random(deg(u)));}
		/** Returns the Gomory-Hu tree of the graph. */
		void gomory_hu_tree (UGRAPH<int,int>& T) const;


		/** Generates a (valid)	empty graph, that is, with zero nodes and zero edges. */
		void gen_empty ();

		/** Generates a sample graph with 5 nodes. */
		void gen_sample ();

		/** Generates a graph equivalent to 'G'. */
		void gen_ugraph (const ugraph& G);

		/** 
			Generates a random graph in the G(n,p) model. The G(n,p) model
			is made of graphs with 'n' nodes where each pair of distinct nodes 
			is joined with an edge with probability 'p'.
		*/		
		void gen_random_np (const int n, const double p);

		/** 
			Generates a random graph in the G(n,m) model. The G(n,m) model
			is made of graphs with 'n' nodes and 'm' edges.
		*/			
		void gen_random_nm (const int n, const int m);

		/** 
			Generates a random linear graph in the RLG(n,c) model. 
			In those graphs, V=[n], and Prob{ij\in E}=c/(j-i+1) for all 1<=i<j<=n.
		*/				
		void gen_random_linear (const int n, const double c=1);	

		/**
			Generates a site percolation graph with parameter 'p'. 
			Some more exaplanation needed!!! 
		*/
		void gen_site_percolation (const double p);	

		/**
			Generates a bond percolation graph with parameter 'p'. 
			Some more exaplanation!!! Subcritical p!!!
		*/
		void gen_bond_percolation (const double p);	

		/**	Generates a complete graph with 'n' nodes. */
		void gen_complete (const int n);

		/**
			Generates a complete bipartite graph with 'a' nodes
			in one side and 'b' nodes on the other side.
		*/
		void gen_complete_bipartite (const int a, const int b);

		/**
			Generates a random bipartite graph with 'a' nodes
			in one side and 'b' nodes on the other side and 'm'
			edges.
		*/
		void gen_random_bipartite_m (const int a, const int b, const int m);

		/**
			Generates a random bipartite graph with 'a' nodes
			in one side and 'b' nodes on the other side and edge probability
			'p' for opsed side nodes.
		*/
		void gen_random_bipartite_p (const int a, const int b, const double p);

		/** Generates a square grid graph of side s. */
		void gen_grid (const int s);

	
	protected:

		/** Returns a reference to the starting adjacency index of a node. */
		inline int& adj (const int u) {return _adj[u];}
		/** Returns a reference to the target node of edge e. */
		inline int& edg (const int e) {return _edg[e];}
		/** Returns a reference to the degree of a node. */
		inline int& deg (const int u) {return _deg[u];}
	
};


#endif

