#include <assert.h>
#include <iostream.h>



typedef struct Node
{
	int id;
	struct Edge *first_edge;	/* in list of incident edges */
	struct Edge *scan_ptr;		/* next edge to be scanned when node will be visited again */
	struct Node *parent;		/* pointer of Gomory-Hu cut tree */
	int mincap;					/* capacity of minimum cut between
								   node and parent in GH cut tree */
	/* subsequent entries for use by maxflow */
	int dist;
	int excess;
	struct Node *bfs_link;		/* for one way BFS working queue */
	struct Node *stack_link;	/* for stack of active nodes */
	bool alive;
	bool unmarked;				/* while BFS in progress */
}
node;

typedef struct Edge
{
	node *adjac;				/* pointer to adjacent node */
	struct Edge *next;			/* in incidence list of node 
								   from which edge is emanating */
	struct Edge *back;			/* pointer to reverse edge */
	int cap;
	int rcap;					/* residual capacity */
}
edge;


typedef struct Graph
{
	int n_nodes;
	node *nodes;
	int n_edges;
	edge *edges;
}
graph;



static node **active;
static int *number;
static int max_dist, bound;
static bool co_check;

  /* This maxflow version is for undirected graphs and
     computes maximum flow only, i.e. the resulting flow
     is not computed for all edges, the graph structure
     is expected to be initialized already, the function
     "init_maxflow" must be called first, then "maxflow"
     may be called any number of times, the function 
     "fini_maxflow" should be called after the final 
     maxflow call.                                      */


bool init_maxflow(int n)
{
	active = new(node *)[n + 1];
	/* holds stacks of active nodes arranged by distances */
	number = new int[n + 1];
	/* counts occurences of node distances in set 
	   of alive nodes, i.e. nodes not contained in
	   the set of nodes disconnected from the sink */
	co_check = true;
	return (true);

}								/* init_maxflow */


void fini_maxflow()
{
	delete[]active;
	delete[]number;
}								/* fini_maxflow */


void global_relabel(graph * gr, node * tptr)
{
	/* breadth first search to get exact distance labels
	   from sink with reordering of stack of active nodes */

	node *front, *rear, *nptr, **ptr;
	edge *eptr;
	int n, level, count, i;

	n = gr->n_nodes;
	for (nptr = &(gr->nodes[n - 1]); nptr >= gr->nodes; nptr--) {
		nptr->unmarked = true;
		nptr->stack_link = 0;
		nptr->scan_ptr = nptr->first_edge;
	}
	tptr->unmarked = false;
	/* initialize stack of active nodes */
	for (ptr = &(active[n]); ptr >= active; ptr--)
		*ptr = 0;
	for (i = 0; i <= n; i++)
		number[i] = 0;
	max_dist = 0;
	count = 1;					/* number of alive nodes */
	front = tptr;
	rear = front;

  bfs_next:
	level = rear->dist + 1;
	eptr = rear->first_edge;
	while (eptr != 0) {
		nptr = eptr->adjac;
		if (nptr->alive && nptr->unmarked && eptr->back->rcap > 0) {
			nptr->unmarked = false;
			nptr->dist = level;
			++count;
			++number[level];
			if (nptr->excess > 0) {
				nptr->stack_link = active[level];
				active[level] = nptr;
				max_dist = level;
			}
			front->bfs_link = nptr;
			front = nptr;
		}
		eptr = eptr->next;
	}
	if (front == rear)
		goto bfs_ready;

	rear = rear->bfs_link;
	goto bfs_next;

  bfs_ready:

	if (count < bound) {		/* identify nodes that are marked alive but have
								   not been reached by BFS and mark them as dead  */
		for (nptr = &(gr->nodes[n - 1]); nptr >= gr->nodes; nptr--)
			if (nptr->unmarked && nptr->alive) {
				nptr->dist = n;
				nptr->alive = false;
			}
		bound = count;
	}
}								/* global_relabel */


int maxflow(graph * gr, node * s_ptr, node * t_ptr)
{
	/* Determines maximum flow and minimum cut between nodes
	   s (= *s_ptr) and t (= *t_ptr) in an undirected graph  

	   References:
	   ----------
	   A. Goldberg/ E. Tarjan: "A New Approach to the
	   Maximum Flow Problem", in Proc. 18th ACM Symp. 
	   on Theory of Computing, 1986.
	 */
	node *aptr, *nptr, *q_front, *q_rear;
	edge *eptr;
	int n, m, level, i, n_discharge;
	int incre;
	int dmin;
	int cap;

	/* node ids range from 1 to n, node array indices  
	   range from 0 to n-1                             */

	n = gr->n_nodes;
	for (nptr = &(gr->nodes[n - 1]); nptr >= gr->nodes; nptr--) {
		nptr->scan_ptr = nptr->first_edge;
		assert(nptr->scan_ptr != 0);	// isolated node in input graph

		nptr->excess = 0;
		nptr->stack_link = 0;
		nptr->alive = true;
		nptr->unmarked = true;
	}
	m = gr->n_edges;
	for (eptr = &(gr->edges[2 * m - 1]); eptr >= gr->edges; eptr--)
		eptr->rcap = eptr->cap;

	for (i = n; i >= 0; i--) {
		number[i] = 0;
		active[i] = 0;
	}
	t_ptr->dist = 0;

	/* breadth first search to get exact distances 
	   from sink and for test of graph connectivity */

	t_ptr->unmarked = false;
	q_front = t_ptr;
	q_rear = q_front;
  bfs_next:
	level = q_rear->dist + 1;
	eptr = q_rear->first_edge;
	while (eptr != 0) {
		if (eptr->adjac->unmarked && eptr->back->rcap > 0) {
			nptr = eptr->adjac;
			nptr->unmarked = false;
			nptr->dist = level;
			++number[level];
			q_front->bfs_link = nptr;
			q_front = nptr;
		}
		eptr = eptr->next;
	}
	if (q_rear == q_front)
		goto bfs_ready;

	q_rear = q_rear->bfs_link;
	goto bfs_next;

  bfs_ready:
	if (co_check) {
		co_check = false;
		for (nptr = &(gr->nodes[n - 1]); nptr >= gr->nodes; --nptr)
			assert(!nptr->unmarked);	// Input graph not connected

	}
	s_ptr->dist = n;			/* number[0] and number[n] not required */
	t_ptr->dist = 0;
	t_ptr->excess = 1;			/* to be subtracted again */


	/* initial preflow push from source node */

	max_dist = 0;				/* = max_dist of active nodes */
	eptr = s_ptr->first_edge;
	while (eptr != 0) {
		nptr = eptr->adjac;
		cap = eptr->rcap;
		nptr->excess += cap;
		s_ptr->excess -= cap;
		eptr->back->rcap += cap;
		eptr->rcap = 0;

		if (nptr != t_ptr && nptr->excess <= cap + 0) {	/* push node nptr onto stack for nptr->dist,
														   but only once in case of int edges     */
			nptr->stack_link = active[nptr->dist];
			active[nptr->dist] = nptr;
			if (nptr->dist > max_dist)
				max_dist = nptr->dist;
		}
		eptr = eptr->next;
	}

	s_ptr->alive = false;
	bound = n - 1;
	n_discharge = 0;

	/* main loop */

	do {						/* get maximum distance active node */
		aptr = active[max_dist];
		while (aptr != 0) {
			active[max_dist] = aptr->stack_link;
			eptr = aptr->scan_ptr;

		  edge_scan:			/* for current active node  */
			nptr = eptr->adjac;
			if (nptr->dist == aptr->dist - 1 && eptr->rcap > 0) {
				incre = aptr->excess;
				if (incre <= eptr->rcap) {	/* perform a non saturating push */
					eptr->rcap -= incre;
					eptr->back->rcap += incre;
					aptr->excess = 0;
					nptr->excess += incre;
					if (nptr->excess <= incre + 0) {	/* push nptr onto active stack */
						nptr->stack_link = active[nptr->dist];
						active[nptr->dist] = nptr;
					}
					aptr->scan_ptr = eptr;
					goto node_ready;
				} else {		/* perform a saturating push */
					incre = eptr->rcap;
					eptr->back->rcap += incre;
					aptr->excess -= incre;
					nptr->excess += incre;
					eptr->rcap = 0;
					if (nptr->excess <= incre + 0) {	/* push nptr onto active stack */
						nptr->stack_link = active[nptr->dist];
						active[nptr->dist] = nptr;
					}
					if (aptr->excess <= 0) {
						aptr->scan_ptr = eptr;
						goto node_ready;
					}
				}
			}
			if (eptr->next == 0) {	/* all incident arcs scanned, but node still
									   has positive excess, check if for all nptr       
									   nptr->dist != aptr->dist                  */

				if (number[aptr->dist] == 1) {	/* put all nodes v with dist[v] >= dist[a] 
												   into the set of "dead" nodes since they
												   are disconnected from the sink          */

					for (nptr = &(gr->nodes[n - 1]);
						 nptr >= gr->nodes; nptr--)
							if (nptr->alive && nptr->dist > aptr->dist) {
							--number[nptr->dist];
							active[nptr->dist] = 0;
							nptr->alive = false;
							nptr->dist = n;
							--bound;
						}
					--number[aptr->dist];
					active[aptr->dist] = 0;
					aptr->alive = false;
					aptr->dist = n;
					--bound;
					goto node_ready;
				} else {		/* determine new label value */
					dmin = n;
					aptr->scan_ptr = 0;
					eptr = aptr->first_edge;
					while (eptr != 0) {
						if (eptr->adjac->dist < dmin && eptr->rcap > 0) {
							dmin = eptr->adjac->dist;
							if (aptr->scan_ptr == 0)
								aptr->scan_ptr = eptr;
						}
						eptr = eptr->next;
					}
					if (++dmin < bound) {	/* ordinary relabel operation */
						--number[aptr->dist];
						aptr->dist = dmin;
						++number[dmin];
						max_dist = dmin;
						eptr = aptr->scan_ptr;
						goto edge_scan;
					} else {
						aptr->alive = false;
						--number[aptr->dist];
						aptr->dist = n;
						--bound;
						goto node_ready;
					}
				}
			} else {
				eptr = eptr->next;
				goto edge_scan;
			}

		  node_ready:
			++n_discharge;
			if (n_discharge == n) {
				n_discharge = 0;
				global_relabel(gr, t_ptr);
			}
			aptr = active[max_dist];
		}						/* aptr != 0 */
		--max_dist;
	}
	while (max_dist > 0);

	return (t_ptr->excess - 1);
}

bool ghc_tree(graph * gr)
{
	/* Determines Gomory/Hu cut tree for input graph with
	   capacitated edges, the tree structures is represented
	   by parent pointers which are part of the node structure,
	   the capacity of a tree edge is stored at the child node,
	   the root of the cut tree is the first node in the list
	   of graph nodes (&gr->nodes[0]). The implementation is
	   described in [1].

	   References:
	   ----------
	   1) D. Gusfield: "Very Simple Algorithms and Programs for
	   All Pairs Network Flow Analysis", Computer Science Di-
	   vision, University of California, Davis, 1987.

	   2) R.E. Gomory and T.C. Hu: "Multi-Terminal Network Flows",
	   SIAM J. Applied Math. 9 (1961), 551-570.

	 */

	node *nptr, *nptr1, *nptrn, *sptr, *tptr;
	edge *eptr;
	int n, m;
	int maxfl;


	n = gr->n_nodes;
	m = gr->n_edges;

	if (!init_maxflow(n))
		return (false);

	nptr1 = gr->nodes;
	nptrn = &(gr->nodes[n - 1]);
	for (nptr = nptrn; nptr >= nptr1; nptr--)
		nptr->parent = nptr1;

	int kkk=0;
	for (sptr = &(gr->nodes[1]); sptr <= nptrn; sptr++) {
		
		if (kkk++==10) {
			cerr << "." << flush;
			kkk=0;
		}
		
		for (eptr = &(gr->edges[m - 1]); eptr >= gr->edges; eptr--)
			eptr->rcap = eptr->cap;
		tptr = sptr->parent;
		maxfl = maxflow(gr, sptr, tptr);
		if (maxfl < 0)
			return (false);

		sptr->mincap = maxfl;
		for (nptr = &(gr->nodes[1]); nptr <= nptrn; nptr++)
			if (nptr != sptr && !nptr->alive && nptr->parent == tptr)
				nptr->parent = sptr;
		if (!tptr->parent->alive) {
			sptr->parent = tptr->parent;
			tptr->parent = sptr;
			sptr->mincap = tptr->mincap;
			tptr->mincap = maxfl;
		}
	}
	fini_maxflow();
cerr << endl;	
	return (true);
}								/* ghc_tree */




bool alloc_graph(int n, int m, graph ** gr)
{
	(*gr) = new graph;
	(*gr)->nodes = new node[n];
	(*gr)->edges = new edge[2 * m];
	return (true);
}


void dealloc_graph(graph * gr)
{
	delete[]gr->nodes;
	delete[]gr->edges;
	delete gr;
}




graph *get_graph(int nodes, int edges, int *u, int *v)
{
	int n, m, i, j, nod1, nod2;
	int cap;
	node *nptr, *nptr1, *nptr2;
	edge *eptr1, *eptr2;
	graph *gr;


	n = nodes;
	m = edges;
	alloc_graph(n, m, &gr);
	gr->n_nodes = n;
	gr->n_edges = m;

	for (i = n, nptr = &(gr->nodes[n - 1]); i > 0; --nptr, --i) {
		nptr->id = i;
		nptr->first_edge = 0;
	}

	eptr1 = &(gr->edges[0]);
	eptr2 = &(gr->edges[m]);
	for (j = 0; j < m; j++) {
		nod1 = u[j] + 1;
		nod2 = v[j] + 1;
		cap = 1;
		/* put edge into internal graph representation */
		--nod1;
		--nod2;
		nptr1 = &(gr->nodes[nod1]);
		nptr2 = &(gr->nodes[nod2]);
		eptr1->adjac = nptr2;
		eptr2->adjac = nptr1;
		eptr1->cap = cap;
		eptr2->cap = cap;
		eptr1->back = eptr2;
		eptr2->back = eptr1;
		if (nptr1->first_edge == 0) {
			nptr1->first_edge = eptr1;
			eptr1->next = 0;
		} else {
			eptr1->next = nptr1->first_edge;
			nptr1->first_edge = eptr1;
		}
		if (nptr2->first_edge == 0) {
			nptr2->first_edge = eptr2;
			eptr2->next = 0;
		} else {
			eptr2->next = nptr2->first_edge;
			nptr2->first_edge = eptr2;
		}
		++eptr1;
		++eptr2;
	}
	return (gr);
}


void get_tree(graph * gr, int *uu, int *vv, int *ff)
{
	node *nptr;
	int c = 0;
	for (nptr = &(gr->nodes[gr->n_nodes - 1]); nptr >= &(gr->nodes[1]);
		 nptr--) {
		uu[c] = nptr->id - 1;
		vv[c] = nptr->parent->id - 1;
		ff[c] = nptr->mincap;
		c++;
	}
}


void gomory_hu(int n, int m, int *u, int *v, int *uu, int *vv, int *ff)
{
	graph *gr;
	gr = get_graph(n, m, u, v);
	ghc_tree(gr);
	get_tree(gr, uu, vv, ff);
	dealloc_graph(gr);
}
